package cloudduggu.com;

import org.apache.http.HttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.spark.api.java.JavaRDD;
import org.apache.spark.api.java.JavaSparkContext;
import org.apache.spark.api.r.RParallelizeServer;
import org.apache.spark.graphx.Edge;
import org.apache.spark.graphx.Graph;
import org.apache.spark.sql.Dataset;
import org.apache.spark.sql.Row;
import org.apache.spark.sql.SparkSession;
import org.apache.spark.storage.StorageLevel;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import scala.Tuple2;
import scala.reflect.ClassTag;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.net.URLEncoder;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

/**
 * Spark GraphX project main code.
 *
 * @author  Sarvesh Kumar (CloudDuggu.com)
 * @version 1.0
 * @since   2020-08-25
 */

public class GraphXCompute {

    static final Logger logger = LoggerFactory.getLogger(GraphXCompute.class);

    public static void main(String[] args) {

        SparkSession spark = SparkSession
                .builder()
                .appName("CloudDuggu Spark GraphX Project")
                .config("spark.master", "local")
                .getOrCreate();

        JavaSparkContext javaSparkContext = new JavaSparkContext(spark.sparkContext());

//		Create an RDD of Employee objects from a csv file
        JavaRDD<Employee> employeeRDD = spark.read()
                .csv(args[0])
                .javaRDD()
                .map(line -> {
                    Employee employee = new Employee();
                    employee.setId(Long.parseLong(line.getString(0)));
                    employee.setPhoto((line.getString(1)));
                    employee.setName(line.getString(2));
                    employee.setBoss(Arrays.asList(line.getString(3).split("-")).stream().map(Long::parseLong).collect(Collectors.toList()));
                    employee.setColleague(Arrays.asList(line.getString(4).split("-")).stream().map(Long::parseLong).collect(Collectors.toList()));
                    employee.setPartner(Arrays.asList(line.getString(5).split("-")).stream().map(Long::parseLong).collect(Collectors.toList()));
                    employee.setFriend(Arrays.asList(line.getString(6).split("-")).stream().map(Long::parseLong).collect(Collectors.toList()));
                    return employee;
                });

//      create row dataset from employee rdd object
        Dataset<Row> employeeDF = spark.createDataFrame(employeeRDD, Employee.class);
        employeeDF.createOrReplaceTempView("employee");

//      create row dataset for all employees id,name values.
        Dataset<Row> allDF = spark.sql("select id, name from employee");


//      start loop for fetching client data and evaluate employees relationship
        int count=0;
        while (true) {

            try {
//              get client request employee id data
                String dataURL= "http://"+args[1]+":"+args[2]+"/data";
                HttpGet dataHttpGet = new HttpGet(dataURL);
                HttpResponse dataResponse = new DefaultHttpClient().execute(dataHttpGet);
                int dataResponseCode = dataResponse.getStatusLine().getStatusCode();

//              if response is success then start evaluation
                if(dataResponseCode == 200) {

                    StringBuilder requestText = new StringBuilder();
                    try(BufferedReader br = new BufferedReader(new InputStreamReader((dataResponse.getEntity().getContent())))) {
                        // Read in all of the post results into a String.
                        for(String line = br.readLine(); line != null; line = br.readLine())
                            requestText.append(line+"\n");
                    }

                    long empId = Long.parseLong(requestText.toString().trim());

                    logger.info("==============================================================");
                    logger.info("["+count+"] Fetch Employee Id : "+empId);
                    logger.info("==============================================================");

//                  get row dataset for selected employee id
                    Dataset<Row> selectDF = spark.sql("select id, name, photo, boss, colleague, partner, friend from employee where id="+empId);

                    List<Tuple2<Object, String>> listOfVertex = new ArrayList<>();
                    List<Edge<String>> listOfEdge = new ArrayList<>();

//                  add vertex values for selected employee
                    listOfVertex.add(new Tuple2<>(empId, allDF.filter("id=" + empId).collectAsList().get(0).getString(1).trim()));

//                  add vertex values for employee - boss relationship
                    selectDF.collectAsList().get(0).getList(3).forEach(b -> {
                        Long bossId = Long.parseLong(b.toString());
                        String bossName = allDF.filter("id=" + b).collectAsList().get(0).getString(1).trim();
                        listOfVertex.add(new Tuple2<>(bossId, bossName));
                        listOfEdge.add(new Edge<>(empId, bossId, "Boss"));
                    });

//                  add vertex values for employee - colleague relationship
                    selectDF.collectAsList().get(0).getList(4).forEach(b -> {
                        Long colleagueId = Long.parseLong(b.toString());
                        String colleagueName = allDF.filter("id=" + b).collectAsList().get(0).getString(1).trim();
                        listOfVertex.add(new Tuple2<>(colleagueId, colleagueName));
                        listOfEdge.add(new Edge<>(empId, colleagueId, "Colleague"));
                    });

//                  add vertex values for employee - partner relationship
                    selectDF.collectAsList().get(0).getList(5).forEach(b -> {
                        Long partnerId = Long.parseLong(b.toString());
                        String partnerName = allDF.filter("id=" + b).collectAsList().get(0).getString(1).trim();
                        listOfVertex.add(new Tuple2<>(partnerId, partnerName));
                        listOfEdge.add(new Edge<>(empId, partnerId, "Partner"));
                    });

//                  add vertex values for employee - friend relationship
                    selectDF.collectAsList().get(0).getList(6).forEach(b -> {
                        Long friendId = Long.parseLong(b.toString());
                        String friendName = allDF.filter("id=" + b).collectAsList().get(0).getString(1).trim();
                        listOfVertex.add(new Tuple2<>(friendId, friendName));
                        listOfEdge.add(new Edge<>(empId, friendId, "Friend"));
                    });

//                  Parallelize created vertex & edge list object into java rdd
                    JavaRDD<Tuple2<Object, String>> vertexRDD = javaSparkContext.parallelize(listOfVertex);
                    JavaRDD<Edge<String>> edgeRDD = javaSparkContext.parallelize(listOfEdge);

                    ClassTag<String> stringTag = scala.reflect.ClassTag$.MODULE$.apply(String.class);

//                  create graph object from vertex and edge rdd
                    Graph<String, String> graph = Graph.apply(
                            vertexRDD.rdd(),
                            edgeRDD.rdd(),
                            "",
                            StorageLevel.MEMORY_ONLY(),
                            StorageLevel.MEMORY_ONLY(),
                            stringTag,
                            stringTag
                    );

                    StringBuilder result = new StringBuilder();

//                  create result string for client from graph vertex and graph edge
                    graph.vertices().toJavaRDD().collect().forEach(v -> result.append(v._1+","+v._2+"\t"));
                    result.append("\n");
                    graph.edges().toJavaRDD().collect().forEach(e -> result.append(e.srcId()+","+e.dstId()+","+e.attr()+"\t"));


                    logger.info("==============================================================");
                    logger.info("["+count+"] ResultString: "+result.toString());
                    logger.info("==============================================================");


//                  Upload Result Output
                    String uploadURL= "http://"+args[1]+":"+args[2]+"/upload?type=result&text=" + URLEncoder.encode( result.toString(), "UTF-8" );

                    HttpPost resultHttpPost = new HttpPost(uploadURL);
                    HttpResponse resultResponse = new DefaultHttpClient().execute(resultHttpPost);
                    int resultResponseCode = resultResponse.getStatusLine().getStatusCode();

                    logger.info("==============================================================");
                    logger.info("["+count+"] Upload Success... : "+resultResponseCode);
                    logger.info("==============================================================");

                }

//              for next lookup wait 5 second
                Thread.sleep(5 * 1000);
            } catch (Exception ex) {
                logger.error("Error: "+ ex.getMessage());
            }

//          count loop value for avoiding infinite process
            count++;
            if(count >=100)
                break;

        }

//      close spark context object after finished client request evaluation.
        javaSparkContext.close();

    }
}
